@@ -1,13 +1,14 @@
 import copy
 import itertools
 import logging
+import spacy
 import warnings
-from types import ModuleType
-from typing import Any, Callable, Dict, List, Optional, Tuple
-
 from functional import seq
 from spacy.matcher import Matcher
 from spacy.tokens import Span
+from spacy.tokens.underscore import get_ext_args
+from types import ModuleType
+from typing import Callable, List, Optional, Tuple
 
 from replacy import default_match_hooks
 from replacy.db import get_forms_lookup, get_match_dict, load_lm
@@ -20,17 +21,58 @@ from replacy.util import (
     get_novel_prop_defaults,
     get_predicates,
     make_doc_if_not_doc,
+    set_known_extensions,
     validate_match_dict,
+    spacy_version
 )
 from replacy.version import __version__
 
-
 logging.basicConfig(level=logging.INFO)
 
-
 PipelineComponent = Callable[[List[Span]], List[Span]]
 
 
+class ESpan(Span):
+    """
+    dangerous version of Span class
+    intentionally bypass the _ attribute so that the class itself has all the properties
+    this can result in name collisions, etc
+
+    Why use it? there are cases where overlapping spans cause problems for the built in spacy.tokens.Span
+    but for some reason this works
+    """
+
+    def __getattribute__(self, name):
+        """
+        when python attempts to access to underscore property, don't let it, give it self
+        this means that:
+
+        ```python
+        >>> doc = nlp("She extracts revenge.")
+        >>> es = ESpan(doc, 1, 2)
+        >>> e._.comment = "yo metaprogramming"
+        >>> e.comment
+        'yo metaprogramming'
+        ```
+        """
+        if name == "_":
+            return self
+        return super().__getattribute__(name)
+
+    @classmethod
+    def set_extension(cls, name, **kwargs):
+        # if we only want to allow default values, this works:
+        default, method, getter, setter = get_ext_args(**kwargs)
+        setattr(cls, name, default)
+        # if we want to allow getters and setters or methods for dynamic props, we have to implement that
+        # I think it is doable using the `property` built-in method as shown here
+        # https://stackoverflow.com/a/1355444/3518108
+
+    @classmethod
+    def has_extension(cls, name):
+        return hasattr(cls, name)
+
+
 class ReplaceMatcher:
     """
     The main unit of functionality. Instantiate with `nlp`, (an instance of spaCy) and a match dict.
@@ -56,19 +98,22 @@ class ReplaceMatcher:
     validate_match_dict = validate_match_dict
 
     def __init__(
-        self,
-        nlp,
-        match_dict=None,
-        forms_lookup=None,
-        custom_match_hooks: Optional[ModuleType] = None,
-        allow_multiple_whitespaces=False,
-        max_suggestions_count=1000,
-        lm_path=None,
-        filter_suggestions=False,
-        default_max_count=None,
-        debug=False,
+            self,
+            nlp,
+            match_dict=None,
+            forms_lookup=None,
+            custom_match_hooks: Optional[ModuleType] = None,
+            allow_multiple_whitespaces=False,
+            max_suggestions_count=1000,
+            lm_path=None,
+            filter_suggestions=False,
+            default_max_count=None,
+            debug=False,
+            SpanClass=Span,
     ):
         self.debug = debug
+        # self.extended_span = extended_span
+        self.Span = SpanClass
         self.logger = logging.getLogger("replaCy")
         self.default_match_hooks = default_match_hooks
         self.custom_match_hooks = custom_match_hooks
@@ -80,10 +125,11 @@ class ReplaceMatcher:
         self.spans: List[Span] = []
         self.max_suggestions_count = max_suggestions_count
         self.forms_lookup = forms_lookup if forms_lookup else get_forms_lookup()
-        self.suggestion_gen = SuggestionGenerator(
-            nlp, forms_lookup, filter_suggestions, default_max_count
+        self.suggestion_gen = SuggestionGenerator(nlp, forms_lookup, filter_suggestions, default_max_count)
+        expected_properties = set_known_extensions(self.Span)
+        self.novel_prop_defaults = get_novel_prop_defaults(
+            self.match_dict, self.Span, expected_properties
         )
-        self.novel_prop_defaults = get_novel_prop_defaults(self.match_dict)
         self._set_scorer(lm_path)
         # Pipeline doesn't include matcher, since doesn't have the signature List[Span] -> None
         self.pipeline: List[Tuple[str, PipelineComponent]] = [
@@ -92,6 +138,10 @@ class ReplaceMatcher:
             ("joiner", join_suggestions),
         ]
 
+    @classmethod
+    def with_espan(cls, *args, **kwargs):
+        return cls(*args, **kwargs, SpanClass=ESpan)
+
     def _init_matcher(self):
         for match_name, ps in self.match_dict.items():
             patterns = copy.deepcopy(ps["patterns"])
@@ -119,6 +169,12 @@ class ReplaceMatcher:
 
             match_hooks = ps.get("match_hook", [])
             callback = self._get_callback(match_name, match_hooks)
+            self._add_matcher_rule(match_name, patterns, callback)
+
+    def _add_matcher_rule(self, match_name, patterns, callback):
+        if spacy_version() >= 3:
+            self.matcher.add(match_name, patterns, on_match=callback, greedy="LONGEST")
+        else:
             self.matcher.add(match_name, callback, patterns)
 
     def _get_callback(self, match_name, match_hooks):
@@ -141,7 +197,7 @@ class ReplaceMatcher:
                 except IndexError:
                     break
             match_name = self.nlp.vocab[match_id].text
-            span = Span(doc, start, end)
+            span = self.Span(doc, start, end)
 
             # find in match_dict if needed
             span._.match_name = match_name
@@ -155,8 +211,6 @@ class ReplaceMatcher:
                     x, doc, start, end, match_name, i
                 )
 
-            span._.description = self.match_dict[match_name].get("description", "")
-            span._.category = self.match_dict[match_name].get("category", "")
             for novel_prop, default_value in self.novel_prop_defaults.items():
                 setattr(
                     span._,
@@ -212,7 +266,7 @@ class ReplaceMatcher:
         return spans
 
     def process_suggestions(
-        self, pre_suggestion, doc, start, end, match_name, pre_suggestion_id
+            self, pre_suggestion, doc, start, end, match_name, pre_suggestion_id
     ):
         # get token <-> pattern correspondence
         pattern = self.match_dict[match_name]["patterns"]
@@ -242,13 +296,13 @@ class ReplaceMatcher:
         return [x[0] for x in self.pipeline]
 
     def add_pipe(
-        self,
-        component: PipelineComponent,
-        name: str = None,
-        before: str = None,
-        after: str = None,
-        first: bool = None,
-        last: bool = None,
+            self,
+            component: PipelineComponent,
+            name: str = None,
+            before: str = None,
+            after: str = None,
+            first: bool = None,
+            last: bool = None,
     ):
         """
         Add a component to the pipeline
@@ -304,6 +358,14 @@ class ReplaceMatcher:
             )
             self.pipeline.append(pipeline_step)
 
+    def remove_pipe(self, name):
+        pipelines = []
+        for p in self.pipeline:
+            if p[0] == name:
+                continue
+            pipelines.append(p)
+        self.pipeline = pipelines
+
     def __call__(self, sent):
         # self.spans must be cleared - global
         self.spans = []
